/* Copyright (c) 1997-1999 Miller Puckette and others.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.TERMS," in this distribution.  */

/* "LT" = comments/mods by Larry Troxler lt@westnet.com */

#include "m_imp.h"
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sched.h>
#include <sys/mman.h>

/* OSS include */

#include <linux/soundcard.h>

/* midi lib include */

#include "s_linux_mididefs.h"
#include "s_linux_midi_io.h"
#include "s_linux_midiparse.h"

/* local function prototypes */

static int oss_open_audio(int,int,int);
static void oss_close_audio( void);
static int oss_send_dacs( void);
static void linux_open_midi( void);
static void oss_calcspace(void);

#ifdef HAVE_ALSA
static int alsa_open_audio(int,int,int);
static void alsa_close_audio( void);
static int alsa_send_dacs( void);
#endif
static void linux_close_midi( void);

/* Defines */
// #define DEBUG(x) {x;}
#define DEBUG(x) x
#define DEBUG2(x) {x;}
// #define DEBUG2(x)

#define DEFAULTCH 2 
#define SAMPLEWIDTH 2
#define BYTESPERCHAN (DACBLKSIZE * SAMPLEWIDTH) 
#define MAXDEV 4    	/* maximum number of input or output devices */
#define OSSCHPERDEV 2	/* channels per OSS device */
#define XFERSAMPS (OSSCHPERDEV * DACBLKSIZE)
#define XFERSIZE (SAMPLEWIDTH * XFERSAMPS)


static int sys_oldaudio = 0;	/* flag to use "fragments" below  */
  /* log base 2 of audio "fragment size" in bytes */
static int sys_logfragsize = 0xb;
static int sys_nfragment = 4;

/* GLOBALS */


int sys_linux_usealsa = 0;

static int cp_sr;
static int sys_inchannels;
static int sys_outchannels;
int sys_schedadvance = 50000; 	/* scheduler advance in microseconds */

static int sys_advance_samples; /* scheduler advance in samples */

t_sample sys_soundout[MAXCH*DACBLKSIZE];
t_sample sys_soundin[MAXCH*DACBLKSIZE];
float sys_dacsr;
int sys_hipriority = 0;

/* our device handles */

typedef struct _dh {
     void* handle;
     int fd;
     int channels;  	/* ALSA only; OSS is always 2ch per device */
     int space; 	/* bytes available for writing/reading  */
     int bufsize;   	/* total buffer size in blocks for this device */
     int dropcount; 	/* number of buffers to drop for resync (output only) */
} t_dac_handle;

static t_dac_handle dac_handles[MAXDEV];
static t_dac_handle adc_handles[MAXDEV];
static int num_outdevices = 0;
static int num_indevices = 0;

static void linux_flush_all_underflows_to_zero(void)
{
/*
    TODO: Implement similar thing for linux (GGeiger) 

    One day we will figure this out, I hope, because it 
    costs CPU time dearly on Intel  - LT
  */
     /*    union fpc_csr f;
	   f.fc_word = get_fpc_csr();
	   f.fc_struct.flush = 1;
	   set_fpc_csr(f.fc_word);
     */
}

static void linux_setchsr(int chin,int chout, int sr)
{
    int nblk;
    sys_inchannels = chin;
    sys_outchannels = chout;
    sys_dacsr = cp_sr = sr;
    sys_advance_samples = (sys_schedadvance * (float)cp_sr) / (1000000.);
    if (sys_advance_samples < 3 * DACBLKSIZE)
    	sys_advance_samples = 3 * DACBLKSIZE;

    post("advance %d samples, samplerate = %d", sys_advance_samples, cp_sr);
    post("inputchannels = %d, outputchannels = %d",
    	sys_inchannels, sys_outchannels);
    memset(sys_soundout, 0, sys_outchannels * (DACBLKSIZE*sizeof(float)));
    memset(sys_soundin, 0, sys_inchannels * (DACBLKSIZE*sizeof(float)));
}

/* public routines */

void sys_open_audio_and_midi(int midiin, int midiout,
    int inchans, int outchans, int rate)
{
    if (rate < 1) rate = 44100;
    if (inchans < 0) inchans = DEFAULTCH;
    if (outchans < 0) outchans = DEFAULTCH;
    linux_flush_all_underflows_to_zero();
#ifdef HAVE_ALSA 
     if (sys_linux_usealsa) alsa_open_audio(inchans,outchans,srate);
     else 
#endif
    oss_open_audio(inchans,outchans,rate);
    linux_open_midi();
}

void sys_close_audio_and_midi(void)
{

#ifdef HAVE_ALSA 
     if (sys_linux_usealsa) alsa_close_audio();
        else
#endif
     oss_close_audio();
     
     linux_close_midi();
}
        

void sys_set_dacs(int chans,  int rate)
{
/* this is unused at the moment. */
}


int sys_send_dacs(void)
{
#ifdef HAVE_ALSA 
     if (sys_linux_usealsa) return alsa_send_dacs();
#endif
     return oss_send_dacs();
}

float sys_getsr(void)
{
     return (sys_dacsr);
}

int sys_get_outchannels(void)
{
     return (sys_outchannels); 
}

int sys_get_inchannels(void) 
{
     return (sys_inchannels);
}

int sys_getchin(void)
{
     return (sys_inchannels);
}


void sys_audiobuf(int n)
{

     /* set the size, in milliseconds, of the audio FIFO */
     if (n < 5) n = 5;
     else if (n > 5000) n = 5000;
     fprintf(stderr, "audio buffer set to %d milliseconds\n", n);
     sys_schedadvance = n * 1000;

}

void sys_reportidle(void)
{
}


/* commandline flags */

void sys_linuxfrags(int n) {
     sys_nfragment = n;
     sys_oldaudio = 1;
     post("fragment count specified; using old-fashioned audio I/O");
}

void sys_linux_alsa(int n) {
     sys_linux_usealsa = 1;
}

void sys_linuxfragsize(int n) {
     sys_logfragsize = n;
     sys_oldaudio = 1;
     post("fragment size specified; using old-fashioned audio I/O");
}

/***************************************************
 *  Code using the OSS API
 */ 


static char ossdsp[] = "/dev/dsp%d"; 

int oss_reset(int fd) {
     int err;
     if ((err = ioctl(fd,SNDCTL_DSP_RESET)) < 0)
	  error("OSS: Could not reset");
     return err;
}

void oss_configure(int dev, int fd,int srate, int dac) {
    int orig, param, nblk;
    audio_buf_info ainfo;
    orig = param = srate; 

    /* samplerate */

    if (ioctl(fd,SNDCTL_DSP_SPEED,&param) == -1)
     fprintf(stderr,"OSS: Could not set sampling rate for device\n");
    else if( orig != param )
    	fprintf(stderr,"OSS: sampling rate: wanted %d, got %d\n",
	    orig, param );				

    // setting the correct samplerate (could be different than expected)     
    srate = param;

    	/* I don't know if you can really do this in general */
    if (sys_oldaudio)
    { 

    	/* setting fragment count and size */

    	param = (sys_nfragment<<16) + sys_logfragsize;

    	post("nfrags %d, fragment %d",
    	    sys_nfragment, sys_logfragsize);

    	if (ioctl(fd,SNDCTL_DSP_SETFRAGMENT,&param) == -1)
    	    fprintf(stderr,"OSS: Could not set fragment size\n");
    }

    /* setting resolution */

    orig = param = AFMT_S16_LE;
    if (ioctl(fd,SNDCTL_DSP_SETFMT,&param) == -1)
    	fprintf(stderr,"OSS: Could not set DSP format\n");
    else if( orig != param )
    	fprintf(stderr,"OSS: DSP format: wanted %d, got %d\n",orig, param );

    /* setting channels */

    orig = param = OSSCHPERDEV;

    if (ioctl(fd,SNDCTL_DSP_CHANNELS,&param) == -1)
    	fprintf(stderr,"OSS: Could not set channels\n");
    else if( orig != param )
    	fprintf(stderr,"OSS: num channels: wanted %d, got %d\n",orig, param );

    if (dac)
    {
	/* use "free space" to learn the buffer size.  Normally you
	should set this to your own desired value; but this seems not
	to be implemented uniformly across different sound cards.  LATER
	we should figure out what to do if the requested scheduler advance
	is greater than this buffer size; for now, we just print something
	out.  */

	if (ioctl(dac_handles[dev].fd, SOUND_PCM_GETOSPACE,&ainfo) < 0)
	   fprintf(stderr,"OSS: ioctl on output device %d failed",dev);
	dac_handles[dev].bufsize = ainfo.bytes;

	if (sys_advance_samples * (SAMPLEWIDTH * OSSCHPERDEV) >
	    dac_handles[dev].bufsize - XFERSIZE)
	{
	    fprintf(stderr,
	    	"OSS: requested audio buffer size %d limited to %d\n",
		    sys_advance_samples * (SAMPLEWIDTH *OSSCHPERDEV),
		    dac_handles[dev].bufsize);
    	    sys_advance_samples =
	    	(dac_handles[dev].bufsize - XFERSAMPS) /
		     (SAMPLEWIDTH *OSSCHPERDEV);
	}
    }
}

int oss_open_audio(int inchans, int outchans,int srate)
{  
    int orig;
    int tmp;
    int inchannels = 0,outchannels = 0;
    char devname[20];
    int i;
    char* buf[SAMPLEWIDTH*OSSCHPERDEV*DACBLKSIZE];
    int num_devs = 0;
    audio_buf_info ainfo;

    num_indevices = num_outdevices = 0;

    /* First check if we can do full duplex */
    /* and open the write ports	*/

    for (num_devs=0; outchannels < outchans; num_devs++) {
	int channels = 2;
	if (num_devs)
	   sprintf(devname, ossdsp, num_devs);
	else
	   sprintf(devname,"/dev/dsp");

	DEBUG(fprintf(stderr,"OSS: device Nr. %d on %s\n",num_outdevices+1,devname);)

	if ((tmp = open(devname,O_WRONLY)) == -1) {
	    DEBUG(fprintf(stderr,"OSS: failed to open %s writeonly\n",
	    	devname);)
	    break;
	}

	if (ioctl(tmp,SNDCTL_DSP_GETCAPS,&orig) == -1)
	    error("OSS: SNDCTL_DSP_GETCAPS failed %s",devname);

	if (inchans > inchannels)
	{
	    if (orig & DSP_CAP_DUPLEX)
	    {
		close(tmp);
		if ((dac_handles[num_outdevices].fd = 
		   adc_handles[num_indevices].fd = 
		   open(devname, O_RDWR)) < 0)
		{
		    fprintf(stderr,
		    "OSS: failed to open dac; continuing without audio\n");
    	    	    linux_setchsr(2,2,44100);
		    return 0;
		}
		else
		{
		    num_outdevices++;
		    num_indevices++;
		    outchannels += channels;
		    inchannels += channels;
		}
	    }
	    else
	    {
		dac_handles[num_outdevices].fd = tmp;
		outchannels += channels;
		num_outdevices++;
	    }
	}
	else
	{
	   dac_handles[num_outdevices].fd = tmp;
	   outchannels += channels;
	   num_outdevices++;
	}
     }

    for (; inchans > inchannels; num_indevices++, num_devs++) {
	int channels = 2;
	post("OSS: requiring extra adc handles");
	if (num_devs)
	   sprintf(devname,ossdsp,num_devs);
	else
	   sprintf(devname,"/dev/dsp");

	if ((tmp = open(devname,O_RDONLY)) == -1)
	{
	   DEBUG(fprintf(stderr,"OSS: failed to open %s read only\n",
		devname);)
	   break;
	}
	else 
	   adc_handles[num_indevices].fd = tmp;
	inchans += channels;
    }

    	/* set sample rate; this also calculates sys_advance_samples.  */
    linux_setchsr(inchannels,outchannels,srate);

    /* configure soundcards */

    for (i=0;i<num_indevices;i++) {
	int param = 0;
	oss_configure(i, adc_handles[i].fd,srate, 0);

	   /* try to mute input mixer */
	ioctl(adc_handles[i].fd,SOUND_MIXER_WRITE_IMIX,&param);
    }
    for (i=0;i<num_outdevices;i++) {
	oss_configure(i, dac_handles[i].fd,srate, 1);
	oss_reset(dac_handles[i].fd);
    }
    
      /* We have to do a read to start the engine. This is 
	 necessary because sys_send_dacs waits until the input
	 buffer is filled and only reads on a filled buffer.
	 This is good, because it's a way to make sure that we
	 will not block */

    if (num_indevices) {
	fprintf(stderr,("OSS: starting engine ... "));
	read(adc_handles[0].fd,buf,SAMPLEWIDTH*OSSCHPERDEV*DACBLKSIZE);
	fprintf(stderr,"done\n");
    }
    return (0);
}

void oss_close_audio( void)
{
     int i;
     for (i=0;i<num_outdevices;i++)
	  close(dac_handles[i].fd);

     for (i=0;i<num_indevices;i++)
          close(adc_handles[i].fd);
}


int sys_dacs_write(int fd,void* buf,long bytes) {
    return write(fd, buf,bytes);
}

int sys_dacs_read(int fd,void*  buf,long bytes) {
     return read(fd, buf,bytes);
}

    /* query audio devices for "available" data size. */
static void oss_calcspace(void)
{
    int dev;
    audio_buf_info ainfo;
    for (dev=0; dev < num_outdevices; dev++)
    {
	if (ioctl(dac_handles[dev].fd, SOUND_PCM_GETOSPACE,&ainfo) < 0)
	   fprintf(stderr,"OSS: ioctl on output device %d failed",dev);
	dac_handles[dev].space = ainfo.bytes;
    }

    for (dev = 0; dev < num_indevices; dev++)
    {
	if (ioctl(adc_handles[dev].fd, SOUND_PCM_GETISPACE,&ainfo) < 0)
	    fprintf(stderr, "OSS: ioctl on input device %d, fd %d failed",
    	    	dev, adc_handles[dev].fd);
	adc_handles[dev].space = ainfo.bytes;
    }
}

void linux_audiostatus(void)
{
    int dev;
    if (!sys_oldaudio)
    {
	oss_calcspace();
	for (dev=0; dev < num_outdevices; dev++)
	    fprintf(stderr, "dac %d space %d\n", dev, dac_handles[dev].space);

	for (dev = 0; dev < num_indevices; dev++)
	    fprintf(stderr, "adc %d space %d\n", dev, adc_handles[dev].space);

	fprintf(stderr, "dac buf %d, advance %d\n", dac_handles[0].bufsize,
    	    sys_advance_samples);
    }
}

/* this call resyncs audio output and input which will cause discontinuities
in audio output and/or input. */ 

static void oss_doresync( void)
{
    int dev, zeroed = 0, wantsize;
    int16_t buf[XFERSAMPS];
    audio_buf_info ainfo;

    	/* 1. if any input devices are ahead (have more than 1 buffer stored),
	    drop one or more buffers worth */
    for (dev = 0; dev < num_indevices; dev++)
    {
    	if (adc_handles[dev].space == 0)
	{
	    sys_dacs_read(adc_handles[dev].fd, buf, XFERSIZE);
	}
	else while (adc_handles[dev].space > XFERSIZE)
	{
    	    sys_dacs_read(adc_handles[dev].fd, buf, XFERSIZE);
	    if (ioctl(adc_handles[dev].fd, SOUND_PCM_GETISPACE, &ainfo) < 0)
	    {
	    	fprintf(stderr, "OSS: ioctl on input device %d, fd %d failed",
    	    	    dev, adc_handles[dev].fd);
	    	break;
	    }
	    adc_handles[dev].space = ainfo.bytes;
	}
    }

    	/* 2. if any output devices are behind, feed them zeros to catch them
	    up */
    for (dev = 0; dev < num_outdevices; dev++)
    {
    	while (dac_handles[dev].space > dac_handles[dev].bufsize - 
	    sys_advance_samples * (OSSCHPERDEV * SAMPLEWIDTH))
	{
    	    if (!zeroed)
	    {
	    	int i;
		for (i = 0; i < XFERSAMPS; i++) buf[i] = 0;
		zeroed = 1;
	    }
    	    sys_dacs_write(dac_handles[dev].fd, buf, XFERSIZE);
	    if (ioctl(dac_handles[dev].fd, SOUND_PCM_GETOSPACE, &ainfo) < 0)
	    {
	    	fprintf(stderr, "OSS: ioctl on output device %d, fd %d failed",
    	    	    dev, dac_handles[dev].fd);
	    	break;
	    }
	    dac_handles[dev].space = ainfo.bytes;
	}
    }
    	/* 3. if any DAC devices are too far ahead, plan to drop the
	    number of frames which will let the others catch up. */
    for (dev = 0; dev < num_outdevices; dev++)
    {
    	if (dac_handles[dev].space > dac_handles[dev].bufsize - 
	    (sys_advance_samples - 1) * (OSSCHPERDEV * SAMPLEWIDTH))
	{
    	    dac_handles[dev].dropcount = sys_advance_samples - 1 - 
	    	(dac_handles[dev].space - dac_handles[dev].bufsize) /
		     (OSSCHPERDEV * SAMPLEWIDTH) ;
	}
	else dac_handles[dev].dropcount = 0;
    }
}

int oss_send_dacs(void)
{
    float *fp1, *fp2;
    long fill;
    int i, j, dev, rtnval = 1;
    int16_t buf[OSSCHPERDEV * DACBLKSIZE], *sp;
    	/* the maximum number of samples we should have in the ADC buffer */
    int idle = 0;
    double timeref, timenow;

    if (!num_indevices && !num_outdevices) return (0);

    if (!sys_oldaudio)
    	oss_calcspace();
    
    	/* determine whether we're idle.  This is true if either (1)
    	some input device has less than one buffer to read or (2) some
	output device has fewer than (sys_advance_samples) blocks buffered
	already. */
    for (dev=0; dev < num_outdevices; dev++)
	if (dac_handles[dev].dropcount ||
	    (dac_handles[dev].bufsize - dac_handles[dev].space >
	    	sys_advance_samples * (SAMPLEWIDTH * OSSCHPERDEV)))
    	    	    idle = 1;
    for (dev=0; dev < num_indevices; dev++)
	if (adc_handles[dev].space < XFERSIZE)
	    idle = 1;

    if (idle && !sys_oldaudio)
    {
    	    /* sometimes---rarely---when the ADC available-byte-count is
	    zero, it's genuine, but usually it's because we're so
	    late that the ADC has overrun its entire kernel buffer.  We
	    distinguish between the two by waiting 2 msec and asking again.
	    There should be an error flag we could check instead; look for this
    	    someday... */
    	for (dev = 0;dev < num_indevices; dev++) 
	    if (adc_handles[dev].space == 0)
	{
    	    audio_buf_info ainfo;
	    sys_microsleep(2000);
	    oss_calcspace();
	    if (adc_handles[dev].space != 0) continue;

    	    	/* here's the bad case.  Give up and resync. */
	    sys_log_error(ERR_ADCSTALL);
	    oss_doresync();
	    return (0);
	}
    	    /* check for slippage between devices, either because
	    data got lost in the driver from a previous late condition, or
	    because the devices aren't synced.  When we're idle, no
	    input device should have more than one buffer readable and
	    no output device should have less than sys_advance_samples-1
	    */
	    
	for (dev=0; dev < num_outdevices; dev++)
	    if (!dac_handles[dev].dropcount &&
		(dac_handles[dev].bufsize - dac_handles[dev].space <
	    	    (sys_advance_samples - 2) * (SAMPLEWIDTH * OSSCHPERDEV)))
    	    		goto badsync;
	for (dev=0; dev < num_indevices; dev++)
	    if (adc_handles[dev].space > 3 * XFERSIZE)
		goto badsync;

	    /* return zero to tell the scheduler we're idle. */
    	return (0);
    badsync:
	sys_log_error(ERR_RESYNC);
	oss_doresync();
	return (0);
    	
    }

	/* do output */

    timeref = sys_getrealtime();
    for (dev=0; dev < num_outdevices; dev++)
    {
    	if (dac_handles[dev].dropcount)
	    dac_handles[dev].dropcount--;
	else
	{
	    for (i = DACBLKSIZE,  fp1 = sys_soundout + OSSCHPERDEV*DACBLKSIZE*dev,
	    	sp = buf; i--; fp1++, sp += OSSCHPERDEV)
	    {
		for (j=0, fp2 = fp1; j<OSSCHPERDEV; j++, fp2 += DACBLKSIZE)
		{
	    		/* clip, don't wrap */
	    	    int s = *fp2 * 32767.;
		    if (s > 32767) s = 32767;
		    else if (s < -32767) s = -32767;
	    	    sp[j] = s;
		}
	    }
	    sys_dacs_write(dac_handles[dev].fd, buf, XFERSIZE);
	    if ((timenow = sys_getrealtime()) - timeref > 0.002)
	    {
	    	if (!sys_oldaudio)
		    sys_log_error(ERR_DACSLEPT);
    	    }
	    timeref = timenow;
    	}
    }
    memset(sys_soundout, 0,
    	num_outdevices * (sizeof(float) * OSSCHPERDEV* DACBLKSIZE));

    	/* do input */

    for (dev = 0; dev < num_indevices; dev++) {

    	sys_dacs_read(adc_handles[dev].fd, buf, XFERSIZE);

	if ((timenow = sys_getrealtime()) - timeref > 0.002)
	{
	    if (!sys_oldaudio)
	    	sys_log_error(ERR_ADCSLEPT);
    	}
	timeref = timenow;

	for (i = DACBLKSIZE,fp1 = sys_soundin + OSSCHPERDEV*DACBLKSIZE*dev,
	    sp = buf; i--; fp1++,  sp += OSSCHPERDEV) {
	   for (j=0;j<sys_inchannels;j++)
		fp1[j*DACBLKSIZE] = (float)sp[j]*3.051850e-05;
	}	  	  
     }	
     return (rtnval);
}


/**************************************************
 *   Code using the ALSA native API
 */ 


#ifdef HAVE_ALSA

/* ALSA includes */

#include <sys/asoundlib.h>
#include <linux/asound.h>

static int num_cards = 0;


#define DEVICE_CHANNELS 2

static int alsa_open_audio(int inchans,int outchans,int srate)
{  
     int card = 0,device, err, i;
     struct snd_ctl_hw_info hw_info;
     void *handle;
     int inchannels = 0;
     int outchannels = 0;
     struct snd_pcm_format pcm_format;
     struct snd_pcm_playback_params pcm_playback_params;
     struct snd_pcm_record_params pcm_record_params;
     struct snd_pcm_playback_status pcm_playback_status;
     struct snd_pcm_record_status pcm_record_status;
     struct snd_pcm_playback_info pcm_playback_info;
     struct snd_pcm_record_info pcm_record_info;

     num_indevices =0;
     num_outdevices = 0;
     if( (num_cards = err = snd_cards()) < 0){
	  fprintf(stderr, "ALSA - ERROR number of cards:  %s\n", snd_strerror(err));
	  return 0;
     } 

     if(num_cards == 0 ){
	  fprintf(stderr, 
		  "WARNING: no soundcards present - continuing without audio\n");
	  return 0;
     } 

     DEBUG(fprintf(stderr, "ALSA - found %d soundcard(s):\n", num_cards));


     /* search for devices */
     for(card=0; card < num_cards; card++){
	  if ( (err = snd_ctl_open( &handle, card )) < 0 ) {
	       fprintf( stderr, "ALSA - ERROR snd_ctl failed: %s\n", snd_strerror(err) );
	       return 0;
	  } 
	  
	  if((err = snd_ctl_hw_info(handle, &hw_info)) < 0){
		    fprintf( stderr, "ALSA - ERROR snd_ctl_hw_info failed: %s\n", snd_strerror(err) );
		    return 0;
	  }
	      
	  DEBUG(fprintf(stderr,"ALSA - DAC: (card: %d / dev: %d / typ: %s / PCMs: %d)\n",
		  card, 0, hw_info.name, hw_info.pcmdevs));

	  for (device = 0;device < hw_info.pcmdevs;device++) {
	       if (inchannels >= inchans && outchannels >= outchans) 
		    goto end_detect;
	       if (outchannels  < sys_outchan) {
		    if( (err = snd_pcm_open(&dac_handles[num_outdevices].handle, card, device,
					    SND_PCM_OPEN_PLAYBACK)) >= 0) { 
			 fprintf(stderr,"ALSA - Card Nr %d, device %d write mode\n",card,device);
			 if((err = snd_pcm_playback_info(dac_handles[num_outdevices].handle, &pcm_playback_info))<0)
			      fprintf(stderr,"ALSA - ERROR pcm_playback: %s\n",snd_strerror(err));
			 outchannels+= pcm_playback_info.max_channels;
			 dac_handles[num_outdevices].channels = pcm_playback_info.max_channels;
			 num_outdevices++;
		    }    
		    else {
		      fprintf(stderr,"ALSA - open write %s\n",snd_strerror(err));
		      sys_outchan = outchannels;
		    }
	       }
	       if (inchannels  < inchans) {
		    if( (err = snd_pcm_open(&adc_handles[num_indevices].handle, card, 0,
					    SND_PCM_OPEN_RECORD)) >= 0) { 
			 fprintf(stderr,"ALSA - Card Nr %d, device %d read mode\n",card,device);
			 if((err=snd_pcm_record_info(adc_handles[card].handle, &pcm_record_info))<0)
			      fprintf(stderr,"ALSA - ERROR pcm_record: %s\n",snd_strerror(err));
			 inchannels+= pcm_record_info.max_channels;
			 adc_handles[num_indevices].channels = pcm_record_info.max_channels;
			 num_indevices++;
		    }    
		    else {
		      fprintf(stderr,"open read %s\n",snd_strerror(err));
		      inchans = inchannels;
		    }
	       }
	  } /* for devices */
     snd_ctl_close(handle);
     } /* for cards */


 end_detect: 
     linux_setchsr(inchans,outchans,srate);

     /* Set the write parameters */

     for (i=0;i<num_outdevices;i++) {
	  pcm_format.format = SND_PCM_SFMT_S16_LE ; /* 16 bit little endian */
	  pcm_format.rate = srate; 
	  pcm_format.channels = dac_handles[i].channels;
	  if (pcm_format.channels > DEVICE_CHANNELS) fprintf(stderr,"ALSA - too much channels foun -- recompile pd with MAXCH adopted");

	  if( (err = snd_pcm_playback_format(dac_handles[i].handle, &pcm_format)) < 0){ 
	       fprintf(stderr,"ERROR pcm_format (DAC) failed: %s\n", snd_strerror(err));
	       return 0;
	  }
	  
	  pcm_playback_params.fragment_size = (1 << sys_logfragsize);
	  pcm_playback_params.fragments_max  = sys_nfragment;
	  pcm_playback_params.fragments_room = 1;
	  
	  if( (err = snd_pcm_playback_params(dac_handles[i].handle,&pcm_playback_params))< 0){ 
	       fprintf(stderr,"ERROR pcm_playback_params failed: %s\n", 
		       snd_strerror(err));
	       return 0;
	  }
	  
     }	       
     
     /* Set the read parameters */

     for (i=0;i<num_indevices;i++) {
	  pcm_format.format = SND_PCM_SFMT_S16_LE ; /* 16 bit little endian */
	  pcm_format.rate = srate; 
	  pcm_format.channels = adc_handles[i].channels;
	       
	  if( (err = snd_pcm_record_format(adc_handles[i].handle, &pcm_format)) < 0){ 
	       fprintf(stderr,"ERROR pcm_format (ADC) failed: %s\n", snd_strerror(err));
	       return 0;
	  }
	  
	  pcm_record_params.fragment_size = (1 << sys_logfragsize);
	  pcm_record_params.fragments_min = 1;
	  
	  if( (err = snd_pcm_record_params(adc_handles[i].handle, &pcm_record_params))< 0){ 
	       fprintf(stderr,"ERROR pcm_record_params failed: %s\n", 
		       snd_strerror(err));
	       return 0;
	       }
	  
	  /* necessary to initialize record */
	  if (adc_handles[i].handle != NULL) {
	       int16_t buf[SAMPLEWIDTH*MAXCH * DACBLKSIZE];
	       snd_pcm_read(adc_handles[i].handle, buf,
			    SAMPLEWIDTH*pcm_format.channels*DACBLKSIZE);
	  }
     }
     return (0);
}


void alsa_close_audio( void)
{
     int i;
     for (i=0;i< num_outdevices;i++)
	  snd_pcm_close(dac_handles[i].handle);
     for (i=0;i< num_indevices;i++)
	  snd_pcm_close(adc_handles[i].handle);
}

//#define DEBUG_WRITE
//#define DEBUG_READ

int alsa_send_dacs(void)
{
     int16_t buf[SAMPLEWIDTH*MAXCH * DACBLKSIZE],*fp2;
     float *fp1;
     int i,j,err;  
     int sampstomove = sys_outchannels * DACBLKSIZE;
     struct snd_pcm_playback_status pcm_playback_status;
     struct snd_pcm_record_status pcm_record_status;
     struct snd_pcm_playback_info pcm_playback_info;
     struct snd_pcm_record_info pcm_record_info;
     int devs=0;
     int outchannels = 0;
     int inchannels = 0;
     int size;

     /* -- write -- */
     for (devs=0;devs<num_outdevices;devs++) {
	  int channels = dac_handles[devs].channels;
	  if((err = snd_pcm_playback_status(dac_handles[devs].handle, &pcm_playback_status)) < 0) {
	       fprintf(stderr,"ERROR pcm_playback_status %s\n",snd_strerror(err));
	       continue;
	  }
	  
	  /*
	  if(pcm_playback_status.count < BYTESPERCHAN*channels){
#ifdef DEBUG_WRITE
	       fprintf(stderr,"\nplayback_status.count %d bytes -> idle\n",
		       pcm_playback_status.count);
#endif 
	       return 0;
	  }
	  */

	  for (i=DACBLKSIZE,fp1=sys_soundout+outchannels*DACBLKSIZE, 
		    fp2 = buf;i--;fp1++,  fp2 += channels) {
	       for (j=0;j<channels;j++)
		    fp2[j] = (int) (32767.0f*fp1[DACBLKSIZE*j]);
	  }	  

#ifdef DEBUG_WRITE
	       fprintf(stderr,"try to write %d bytes of %d ...",
		       BYTESPERCHAN*channels,pcm_playback_status.count);
#endif
	  size = snd_pcm_write(dac_handles[devs].handle,buf,BYTESPERCHAN*channels);
#ifdef DEBUG_WRITE
	       fprintf(stderr,"%d written\n",
		       BYTESPERCHAN*channels,size);
#endif
	  outchannels+=channels;
     }

     /* -- read -- */

     for (devs=0;devs<num_indevices;devs++) {
          int border = sys_inchannels*BYTESPERCHAN*(cp_advance+1);
	  int channels = adc_handles[devs].channels;

	  if((err = snd_pcm_record_status(adc_handles[devs].handle, &pcm_record_status)) < 0)
	       continue;

	  
	  if( pcm_record_status.count > border){
	       while(pcm_record_status.count > border){
#ifdef DEBUG_READ
		    fprintf(stderr,"byte = %d, border = %d \n",
			    pcm_record_status.count,border);
#endif
//		    post("drop ADC buf");
		    if((err = snd_pcm_record_status(adc_handles[devs].handle, &pcm_record_status)) < 0){
			 fprintf(stderr,"ERROR pcm_record_status failed: %s\n",
				 snd_strerror(err));
			 return 0;
		    }
		    snd_pcm_read(adc_handles[devs].handle,buf,BYTESPERCHAN*channels);
	       }
	       
	  }
	  if ( pcm_record_status.count >= BYTESPERCHAN*channels){
#ifdef DEBUG_READ
	       fprintf(stderr,"try to read %d bytes of %d ...",
		       BYTESPERCHAN*channels,pcm_record_status.count);
#endif
	       size = snd_pcm_read(adc_handles[devs].handle,buf,BYTESPERCHAN*channels);
#ifdef DEBUG_READ
	       fprintf(stderr,"%d read\n",size);
#endif
	  } else{
#ifdef DEBUG_READ
	       fprintf(stderr,"only %d bytes ready to read\n",pcm_record_status.count );
#endif
	       memset(buf, 0, BYTESPERCHAN*channels);
	  } 
     
	  for (i = DACBLKSIZE,fp1 = sys_soundin+inchannels*DACBLKSIZE,
		    fp2 = buf; i--;fp1++,  fp2 += channels) {
	       for (j=0;j<channels;j++)
		    fp1[j*DACBLKSIZE] = (float)fp2[j]/32767.;
	       /*3.051850e-05;*/
	  }
	  inchannels+=channels;
     }
     memset(sys_soundout, 0, sampstomove*sizeof(float));
     return (1);
}

#endif /* HAVE_ALSA */

/* ------------------------- MIDI -------------------------- */
#define NPORTS 16

void inmidi_rtin(int portno, int cmd);
void inmidi_mtc(int portno, int frame, int sec, int min, int hr);
void inmidi_noteon(int portno, int channel, int pitch, int velo);
void inmidi_controlchange(int portno, int channel, int ctlnumber, int value);
void inmidi_programchange(int portno, int channel, int value);
void inmidi_pitchbend(int portno, int channel, int value);
void inmidi_aftertouch(int portno, int channel, int value);
void inmidi_polyaftertouch(int portno, int channel, int pitch, int value);

static int  linux_nports;
MIDIParser *mpa[NPORTS];
MIDIPort *midiport[NPORTS];

void inmidi_noteonoff(int portno, int channel, int pitch, int velon,int veloff)
{
     /* ignore veloff, pd does not respect it */
     inmidi_noteon(portno, channel, pitch, velon);
}

void linux_open_midi(void)
{
     int i;
/* Look for MIDI devices */
     MIDIPort *mp;

     linux_nports = 0;

  
     if(midi_open()<=0)goto bad;
  
     for (mp=midi_getfirstport();linux_nports<NPORTS && mp != NULL;mp=midi_getnextport()){

	  midiport[linux_nports] = mp;

	  mpa[linux_nports] = mp_parse_init(linux_nports);  /* .... */

	  /* initialize all reaction functions */
	  mpa[linux_nports]->noteonoff = inmidi_noteonoff;
	  mpa[linux_nports]->noteon = NULL;
	  mpa[linux_nports]->noteoff = NULL;

	  mpa[linux_nports]->controlchange = inmidi_controlchange;
	  mpa[linux_nports]->programchange = inmidi_programchange;
	  mpa[linux_nports]->pitchbend = inmidi_pitchbend;
	  mpa[linux_nports]->aftertouch = inmidi_aftertouch;
	  mpa[linux_nports]->polyaftertouch = inmidi_polyaftertouch;
	  /* system messages */
	  mpa[linux_nports]->mode = NULL;
	  mpa[linux_nports]->quarterframe = NULL; /* inmidi_mtc */
	  mpa[linux_nports]->mode = NULL;
	  mpa[linux_nports]->songposition = NULL;
	  mpa[linux_nports]->songselect = NULL;
	  mpa[linux_nports]->tunerequest = NULL;
	  mpa[linux_nports]->sysex = NULL;
	  mpa[linux_nports]->realtime = NULL; /* inmidi_rtin; */


	  linux_nports++;
     };
  

     if (linux_nports == 1) 
	  fprintf(stderr,"MIDIlib: Found one MIDI port on %s\n",midiport[0]->devname);
     else 
	  for(i=0;i<linux_nports;i++) 
	       fprintf(stderr,"MIDIlib Found MIDI port %d on %s\n",i,midiport[i]->devname);
			
     return;

 bad:
     error("failed to open MIDI ports; continuing without MIDI.");
     linux_nports = 0;
     return;
}


/* not used but should... */
void linux_close_midi()
{
     int i;
     for(i=0;i<linux_nports;i++){
	  mp_parse_exit(mpa[i]);
	 
     }

     midi_close();
}


void sys_putmidimess(int portno, int a, int b, int c)
{
  if (linux_nports) {
       switch (md_msglen(a)){
       case 2:
	    midi_out(midiport[portno],a);	 
	    midi_out(midiport[portno],b);	 
	    midi_out(midiport[portno],c);
	    return;
       case 1:
	    midi_out(midiport[portno],a);	 
	    midi_out(midiport[portno],b);	 
	    return;
       case 0:
	    midi_out(midiport[portno],a);	 
	    return;
       };
  }
}


void sys_poll_midi(void)
{
     int i;


     for (i = 0; i < linux_nports; i++)
     {
	  if(midi_instat(midiport[i]))
	       mp_parse(mpa[i],midi_in(midiport[i]));
     }
}


/* ------------- HELPS ---------------- */

void sys_set_priority( void) 
{
    struct sched_param par;
    int p1,p2, success = 0;
#if defined(_POSIX_PRIORITY_SCHEDULING)

    p1 = sched_get_priority_min(SCHED_FIFO);
    p2 = sched_get_priority_max(SCHED_FIFO);

    par.sched_priority = 99; /* Obviously this needs to be refined - LT */

    if (sched_setscheduler(0,SCHED_FIFO,&par) != -1)
       post("warning: high priority scheduling enabled");
    else
    	sys_hipriority = 1;
#endif

#ifdef _POSIX_MEMLOCK
    if (mlockall(MCL_FUTURE) != -1) 
    	post("warning: memory locking enabled");
#endif
    

}





